// This is a cut-down version of the same file found in the node-oracledb
// driver located at https://github.com/oracle/node-oracledb; this file is
// solely for the purpose of testing Node issue 28988

#include "njsModule.h"

// class methods
static NJS_NAPI_METHOD(njsConnection_getDbObjectClass);

// asynchronous methods
static NJS_ASYNC_METHOD(njsConnection_getDbObjectClassAsync);

// post asynchronous methods
static NJS_ASYNC_POST_METHOD(njsConnection_getDbObjectClassPostAsync);

// processing arguments methods
static NJS_PROCESS_ARGS_METHOD(njsConnection_getDbObjectClassProcessArgs);

// finalize
static NJS_NAPI_FINALIZE(njsConnection_finalize);

// properties defined by the class
static const napi_property_descriptor njsClassProperties[] = {
    { "_getDbObjectClass", NULL, njsConnection_getDbObjectClass, NULL, NULL,
            NULL, napi_default, NULL },
    { NULL, NULL, NULL, NULL, NULL, NULL, napi_default, NULL }
};

// class definition
const njsClassDef njsClassDefConnection = {
    "Connection", sizeof(njsConnection), njsConnection_finalize,
    njsClassProperties, NULL, false
};


//-----------------------------------------------------------------------------
// njsConnection_createBaton()
//   Create the baton used for asynchronous methods and initialize all
// values. The connection is also checked to see if it is open. If this fails
// for some reason, an exception is thrown.
//-----------------------------------------------------------------------------
bool njsConnection_createBaton(napi_env env, napi_callback_info info,
        size_t numArgs, napi_value *args, njsBaton **baton)
{
    njsConnection *conn;
    njsBaton *tempBaton;

    if (!njsUtils_createBaton(env, info, numArgs, args, &tempBaton))
        return false;
    conn = (njsConnection*) tempBaton->callingInstance;
    tempBaton->oracleDb = conn->oracleDb;

    *baton = tempBaton;
    return true;
}


//-----------------------------------------------------------------------------
// njsConnection_finalize()
//   Invoked when the njsConnection object is garbage collected.
//-----------------------------------------------------------------------------
static void njsConnection_finalize(napi_env env, void *finalizeData,
        void *finalizeHint)
{
    njsConnection *conn = (njsConnection*) finalizeData;

    NJS_FREE_AND_CLEAR(conn->tag);
    free(conn);
}


//-----------------------------------------------------------------------------
// njsConnection_getDbObjectClass()
//   Looks up a database object type given its name and returns it to the
// caller.
//
// PARAMETERS
//   - name
//   - JS callback which will receive (error, cls)
//-----------------------------------------------------------------------------
static napi_value njsConnection_getDbObjectClass(napi_env env,
        napi_callback_info info)
{
    napi_value args[2];
    njsBaton *baton;

    if (!njsConnection_createBaton(env, info, 2, args, &baton))
        return NULL;
    if (!njsConnection_getDbObjectClassProcessArgs(baton, env, args)) {
        njsBaton_reportError(baton, env);
        return NULL;
    }
    njsBaton_queueWork(baton, env, "GetDbObjectClass",
            njsConnection_getDbObjectClassAsync,
            njsConnection_getDbObjectClassPostAsync, 2);
    return NULL;
}


//-----------------------------------------------------------------------------
// njsConnection_getDbObjectClassAsync()
//   Worker function for njsConnection_getDbObjectClass().
//-----------------------------------------------------------------------------
static bool njsConnection_getDbObjectClassAsync(njsBaton *baton)
{
    return true;
}


//-----------------------------------------------------------------------------
// njsConnection_getDbObjectClassPostAsync()
//   Sets up the arguments for the callback to JS.
//-----------------------------------------------------------------------------
static bool njsConnection_getDbObjectClassPostAsync(njsBaton *baton,
        napi_env env, napi_value *args)
{
    njsDbObjectType *objType;

    return njsDbObject_getSubClass(baton, env, &args[1], &objType);
}


//-----------------------------------------------------------------------------
// njsConnection_getDbObjectClassProcessArgs()
//   Processes the arguments provided by the caller and place them on the
// baton.
//-----------------------------------------------------------------------------
static bool njsConnection_getDbObjectClassProcessArgs(njsBaton *baton,
        napi_env env, napi_value *args)
{
    if (!njsUtils_getStringArg(env, args, 0, &baton->name, &baton->nameLength))
        return false;

    return true;
}


//-----------------------------------------------------------------------------
// njsConnection_newFromBaton()
//   Called when a connection is being created from the baton.
//-----------------------------------------------------------------------------
bool njsConnection_newFromBaton(njsBaton *baton, napi_env env,
        napi_value *connObj)
{
    njsConnection *conn;

    // create new instance
    if (!njsUtils_genericNew(env, &njsClassDefConnection,
            baton->oracleDb->jsConnectionConstructor, connObj,
            (njsBaseInstance**) &conn))
        return false;

    // transfer the requested tag to the new object
    if (baton->tagLength > 0) {
        conn->retag = true;
        conn->tag = baton->tag;
        conn->tagLength = baton->tagLength;
        baton->tag = NULL;
        baton->tagLength = 0;
    }

    // copy the oracleDb instance to the new object
    conn->oracleDb = baton->oracleDb;

    return true;
}
