// This is a cut-down version of the same file found in the node-oracledb
// driver located at https://github.com/oracle/node-oracledb; this file is
// solely for the purpose of testing Node issue 28988

#include "njsModule.h"

// class methods
static NJS_NAPI_METHOD(njsDbObject_copy);

// finalize
static NJS_NAPI_FINALIZE(njsDbObject_finalize);
static NJS_NAPI_FINALIZE(njsDbObjectType_finalize);

// other methods used internally
static bool njsDbObjectType_populate(njsDbObjectType *objType, napi_env env,
        napi_value jsObjectType, njsBaton *baton);
static bool njsDbObject_wrap(napi_env env, napi_value value,
        njsDbObject **obj);


//-----------------------------------------------------------------------------
// njsDbObject_copy()
//   Create a copy of the object and return that. The copy is independent of
// the original object that was copied.
//-----------------------------------------------------------------------------
static napi_value njsDbObject_copy(napi_env env, napi_callback_info info)
{
    napi_value thisArg, returnValue = NULL;
    njsDbObjectType *objType = NULL;
    size_t actualArgs = 0;
    njsDbObject *obj;

    // get callback information and validate the number of arguments
    if (napi_get_cb_info(env, info, &actualArgs, NULL, &thisArg,
            (void**) &objType) != napi_ok) {
        njsUtils_genericThrowError(env);
        return NULL;
    }
    if (actualArgs != 0) {
        njsUtils_throwError(env, errInvalidNumberOfParameters);
        return NULL;
    }

    // get instance and create copy
    if (!njsDbObject_getInstance(objType->oracleDb, env, thisArg, &obj))
        return NULL;
    njsDbObject_new(obj->type, env, &returnValue);
    return returnValue;
}


//-----------------------------------------------------------------------------
// njsDbObject_finalize()
//   Invoked when njsDbObject is garbage collected.
//-----------------------------------------------------------------------------
static void njsDbObject_finalize(napi_env env, void *finalizeData,
        void *finalizeHint)
{
    njsDbObject *obj = (njsDbObject*) finalizeData;

    free(obj);
}


//-----------------------------------------------------------------------------
//  njsDbObject_getInstance()
//    Returns the instance associated with the object. Since objects may be
// created in JavaScript, an instance may not be associated at all. In that
// case, a new instance is created and associated with the object.
//-----------------------------------------------------------------------------
bool njsDbObject_getInstance(njsOracleDb *oracleDb, napi_env env,
        napi_value value, njsDbObject **obj)
{
    njsDbObject *tempObj;

    // if JavaScript object is wrapped, instance is returned immediately
    if (napi_unwrap(env, value, (void**) obj) == napi_ok)
        return true;

    // perform wrap, if needed
    if (!njsDbObject_wrap(env, value, &tempObj))
        return false;

    *obj = tempObj;
    return true;
}


//-----------------------------------------------------------------------------
//  njsDbObject_getSubClass()
//    Gets the sub class for the specified type.
//-----------------------------------------------------------------------------
bool njsDbObject_getSubClass(njsBaton *baton, napi_env env, napi_value *cls,
        njsDbObjectType **objectType)
{
    napi_value fn, args[2], callingObj, prototype;
    njsDbObjectType *tempObjectType;

    // call into JavaScript to get sub class (stored in cache on connection)
    NJS_CHECK_NAPI(env, napi_create_string_utf8(env, "NODEJS",
            NAPI_AUTO_LENGTH, &args[0]))
    NJS_CHECK_NAPI(env, napi_create_string_utf8(env, baton->name,
            baton->nameLength, &args[1]))
    NJS_CHECK_NAPI(env, napi_get_reference_value(env, baton->jsCallingObj,
            &callingObj))
    NJS_CHECK_NAPI(env, napi_get_named_property(env, callingObj,
            "_getDbObjectClassJS", &fn))
    NJS_CHECK_NAPI(env, napi_call_function(env, callingObj, fn, 2, args, cls))

    // get the class prototype; if it has already been wrapped, it has been
    // fully populated and there is no need to do anything further
    NJS_CHECK_NAPI(env, napi_get_named_property(env, *cls, "prototype",
            &prototype))
    if (napi_unwrap(env, prototype, (void**) objectType) == napi_ok)
        return true;

    // allocate memory for structure; memory is zero-ed
    tempObjectType = (njsDbObjectType*) calloc(1, sizeof(njsDbObjectType));
    if (!tempObjectType)
        return njsUtils_throwError(env, errInsufficientMemory);

    // populate object type (C) and prototype (JS) with all information about
    // the object type
    if (!njsDbObjectType_populate(tempObjectType, env, prototype, baton)) {
        njsDbObjectType_finalize(env, tempObjectType, NULL);
        return false;
    }

    // wrap the structure for use by JavaScript
    if (napi_wrap(env, prototype, tempObjectType, njsDbObjectType_finalize,
            NULL, NULL) != napi_ok) {
        njsUtils_genericThrowError(env);
        njsDbObjectType_finalize(env, tempObjectType, NULL);
        return false;
    }

    *objectType = tempObjectType;
    return true;
}


//-----------------------------------------------------------------------------
//  njsDbObject_new()
//    To create a new instance of njsDbObject
//-----------------------------------------------------------------------------
bool njsDbObject_new(njsDbObjectType *objType, napi_env env, napi_value *value)
{
    napi_value constructor;
    njsDbObject *obj;

    // create instance and wrap it
    NJS_CHECK_NAPI(env, napi_get_reference_value(env,
            objType->jsDbObjectConstructor, &constructor))
    NJS_CHECK_NAPI(env, napi_new_instance(env, constructor, 0, NULL, value))
    if (!njsDbObject_wrap(env, *value, &obj))
        return false;

    return true;
}


//-----------------------------------------------------------------------------
// njsDbObject_wrap()
//   Wraps the specified value with a new object instance and performs basic
// initialization, if needed. For collections which are proxied, the proxy
// target may already have been wrapped, so check for that first before
// creating a new instance.
//-----------------------------------------------------------------------------
static bool njsDbObject_wrap(napi_env env, napi_value value, njsDbObject **obj)
{
    napi_value prototype, temp;
    njsDbObjectType *objType;
    njsDbObject *tempObj;

    // acquire the prototype from the value; if the prototype does not wrap the
    // object type, then we are dealing with the proxy; in that case, acquire
    // the special property _target which will return the target of the proxy
    // and acquire the prototype from that value instead
    NJS_CHECK_NAPI(env, napi_get_prototype(env, value, &prototype))
    if (napi_unwrap(env, prototype, (void**) &objType) != napi_ok) {
        NJS_CHECK_NAPI(env, napi_get_named_property(env, value, "_target",
                &temp))
        value = temp;
        if (napi_unwrap(env, value, (void**) obj) == napi_ok)
            return true;
        NJS_CHECK_NAPI(env, napi_get_prototype(env, value, &prototype))
        NJS_CHECK_NAPI(env, napi_unwrap(env, prototype, (void**) &objType))
    }

    // create a new object instance
    tempObj = calloc(1, sizeof(njsDbObject));
    if (!tempObj)
        return njsUtils_throwError(env, errInsufficientMemory);
    tempObj->type = objType;
    if (napi_wrap(env, value, tempObj, njsDbObject_finalize, NULL,
            NULL) != napi_ok) {
        free(tempObj);
        return njsUtils_genericThrowError(env);
    }

    *obj = tempObj;
    return true;
}


//-----------------------------------------------------------------------------
// njsDbObjectType_finalize()
//   Invoked when njsDbObjectType is garbage collected.
//-----------------------------------------------------------------------------
static void njsDbObjectType_finalize(napi_env env, void *finalizeData,
        void *finalizeHint)
{
    njsDbObjectType *type = (njsDbObjectType*) finalizeData;

    NJS_DELETE_REF_AND_CLEAR(type->jsDbObjectConstructor);
    NJS_FREE_AND_CLEAR(type->descriptors);
    NJS_FREE_AND_CLEAR(type->fqn);
    free(type);
}


//-----------------------------------------------------------------------------
// njsDbObjectType_getFromClass()
//   Acquires the object type from a class. The object type instance is stored
// on the prototype so a new instance needs to be created, the prototype
// acquired from that new instance, and then the object type instance
// unwrapped. There is a slight complication, however, in that collections are
// proxied, so we need to acquire the actual target first.
//-----------------------------------------------------------------------------
bool njsDbObjectType_getFromClass(napi_env env, napi_value cls,
        njsDbObjectType **objType)
{
    napi_value prototype, temp, tempObj;

    NJS_CHECK_NAPI(env, napi_new_instance(env, cls, 0, NULL, &tempObj))
    NJS_CHECK_NAPI(env, napi_get_prototype(env, tempObj, &prototype))
    if (napi_unwrap(env, prototype, (void**) objType) != napi_ok) {
        NJS_CHECK_NAPI(env, napi_get_named_property(env, tempObj, "_target",
                &temp))
        NJS_CHECK_NAPI(env, napi_get_prototype(env, temp, &prototype))
        NJS_CHECK_NAPI(env, napi_unwrap(env, prototype, (void**) objType))
    }
    return true;
}


//-----------------------------------------------------------------------------
// njsDbObjectType_populate()
//   Populates the object type and its JS equivalent with information about the
// object type.
//-----------------------------------------------------------------------------
static bool njsDbObjectType_populate(njsDbObjectType *objType, napi_env env,
        napi_value jsObjectType, njsBaton *baton)
{
    size_t numProperties;
    napi_value temp;

    // transfer basic data to instance
    objType->oracleDb = baton->oracleDb;

    // keep a reference to the constructor
    NJS_CHECK_NAPI(env, napi_get_named_property(env, jsObjectType,
            "constructor", &temp))
    NJS_CHECK_NAPI(env, napi_create_reference(env, temp, 1,
            &objType->jsDbObjectConstructor))

    numProperties = 1;
    objType->descriptors = calloc(numProperties,
                sizeof(napi_property_descriptor));
    if (!objType->descriptors)
        return njsBaton_setError(baton, errInsufficientMemory);
    objType->descriptors[0].utf8name = "copy";
    objType->descriptors[0].method = njsDbObject_copy;
    objType->descriptors[0].data = objType;
    objType->descriptors[0].attributes = napi_configurable;
    NJS_CHECK_NAPI(env, napi_define_properties(env, jsObjectType,
            numProperties, objType->descriptors))

    return true;
}
