// This is a cut-down version of the same file found in the node-oracledb
// driver located at https://github.com/oracle/node-oracledb; this file is
// solely for the purpose of testing Node issue 28988

#include "njsModule.h"

// class methods
static NJS_NAPI_METHOD(njsOracleDb_getConnection);

// asynchronous methods
static NJS_ASYNC_METHOD(njsOracleDb_getConnectionAsync);

// post asynchronous methods
static NJS_ASYNC_POST_METHOD(njsOracleDb_getConnectionPostAsync);

// processing arguments methods
static NJS_PROCESS_ARGS_METHOD(njsOracleDb_getConnectionProcessArgs);

// finalize
static NJS_NAPI_FINALIZE(njsOracleDb_finalize);

// define constants exposed to JS
static njsConstant njsClassConstants[] = {

    // bind directions
    { "BIND_IN", NJS_BIND_IN },
    { "BIND_INOUT", NJS_BIND_INOUT },
    { "BIND_OUT", NJS_BIND_OUT },

    // outFormat values
    { "OUT_FORMAT_ARRAY", NJS_ROWS_ARRAY },
    { "OUT_FORMAT_OBJECT", NJS_ROWS_OBJECT },
    { "ARRAY", NJS_ROWS_ARRAY },
    { "OBJECT", NJS_ROWS_OBJECT },

    // terminal
    { NULL, 0 }
};

// properties defined by the class
static const napi_property_descriptor njsClassProperties[] = {
    { "_getConnection", NULL, njsOracleDb_getConnection, NULL, NULL, NULL,
            napi_default, NULL },
    { NULL, NULL, NULL, NULL, NULL, NULL, napi_default, NULL }
};

// class definition
const njsClassDef njsClassDefOracleDb = {
    "OracleDb", sizeof(njsOracleDb), NULL, njsClassProperties,
    njsClassConstants, false
};


//-----------------------------------------------------------------------------
// njsOracleDb_finalize()
//   Invoked when the njsOracleDb object is garbage collected.
//-----------------------------------------------------------------------------
static void njsOracleDb_finalize(napi_env env, void *finalizeData,
        void *finalizeHint)
{
    njsOracleDb *oracleDb = (njsOracleDb*) finalizeData;

    NJS_FREE_AND_CLEAR(oracleDb->connectionClass);
    NJS_FREE_AND_CLEAR(oracleDb->edition);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsBaseDbObjectConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsDateConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsConnectionConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsLobConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsPoolConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsResultSetConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsSodaCollectionConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsSodaDatabaseConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsSodaDocCursorConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsSodaDocumentConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsSodaOperationConstructor);
    NJS_DELETE_REF_AND_CLEAR(oracleDb->jsSubscriptions);
    free(oracleDb);
}


//-----------------------------------------------------------------------------
// njsOracleDb_getConnection()
//   Create a standalone connection to the database.
//
// PARAMETERS
//   - options
//   - JS callback which will receive (error, pool)
//-----------------------------------------------------------------------------
static napi_value njsOracleDb_getConnection(napi_env env,
        napi_callback_info info)
{
    napi_value args[2];
    njsBaton *baton;

    // verify number of arguments and create baton
    if (!njsUtils_createBaton(env, info, 2, args, &baton))
        return NULL;
    baton->oracleDb = (njsOracleDb*) baton->callingInstance;

    // get information from arguments and store on the baton
    if (!njsOracleDb_getConnectionProcessArgs(baton, env, args)) {
        njsBaton_reportError(baton, env);
        return NULL;
    }

    // queue work
    if (!njsBaton_queueWork(baton, env, "GetConnection",
            njsOracleDb_getConnectionAsync,
            njsOracleDb_getConnectionPostAsync, 2))
        return NULL;

    return NULL;
}


//-----------------------------------------------------------------------------
// njsOracleDb_getConnectionAsync()
//   Worker function for GetConnection() performed on thread. This establishes
// the connection using the information found in the baton.
//-----------------------------------------------------------------------------
static bool njsOracleDb_getConnectionAsync(njsBaton *baton)
{
    return true;
}


//-----------------------------------------------------------------------------
// njsOracleDb_getConnectionPostAsync()
//   Sets up the arguments for the callback to JS. The connection object is
// created and passed as the second argument. The first argument is always the
// error and at this point it is known that no error has taken place.
//-----------------------------------------------------------------------------
static bool njsOracleDb_getConnectionPostAsync(njsBaton *baton, napi_env env,
        napi_value *args)
{
    return njsConnection_newFromBaton(baton, env, &args[1]);
}


//-----------------------------------------------------------------------------
// njsOracleDb_getConnectionProcessArgs()
//   Process the arguments for the GetConnection() call.
//-----------------------------------------------------------------------------
static bool njsOracleDb_getConnectionProcessArgs(njsBaton *baton,
        napi_env env, napi_value *args)
{
    bool found;

    // copy items used from the OracleDb class since they may change after
    // the asynchronous function begins
    baton->stmtCacheSize = baton->oracleDb->stmtCacheSize;
    baton->externalAuth = baton->oracleDb->externalAuth;
    baton->events = baton->oracleDb->events;
    if (!njsUtils_copyString(env, baton->oracleDb->connectionClass,
            baton->oracleDb->connectionClassLength, &baton->connectionClass,
            &baton->connectionClassLength))
        return false;
    if (!njsUtils_copyString(env, baton->oracleDb->edition,
            baton->oracleDb->editionLength, &baton->edition,
            &baton->editionLength))
        return false;

    // check the various options
    if (!njsBaton_getStringFromArg(baton, env, args, 0, "user", &baton->user,
            &baton->userLength, NULL))
        return false;
    if (!njsBaton_getStringFromArg(baton, env, args, 0, "password",
            &baton->password, &baton->passwordLength, NULL))
        return false;
    if (!njsBaton_getStringFromArg(baton, env, args, 0, "connectString",
            &baton->connectString, &baton->connectStringLength, &found))
        return false;
    if (!found && !njsBaton_getStringFromArg(baton, env, args, 0,
            "connectionString", &baton->connectString,
            &baton->connectStringLength, NULL))
        return false;
    if (!njsBaton_getStringFromArg(baton, env, args, 0, "newPassword",
            &baton->newPassword, &baton->newPasswordLength, NULL))
        return false;
    if (!njsBaton_getStringFromArg(baton, env, args, 0, "edition",
            &baton->edition, &baton->editionLength, NULL))
        return false;
    if (!njsBaton_getUnsignedIntFromArg(baton, env, args, 0, "stmtCacheSize",
            &baton->stmtCacheSize, NULL))
        return false;
    if (!njsBaton_getUnsignedIntFromArg(baton, env, args, 0, "privilege",
            &baton->privilege, NULL))
        return false;
    if (!njsBaton_getBoolFromArg(baton, env, args, 0, "externalAuth",
            &baton->externalAuth, NULL))
        return false;
    if (!njsBaton_getBoolFromArg(baton, env, args, 0, "events", &baton->events,
            NULL))
        return false;

    return true;
}


//-----------------------------------------------------------------------------
// njsOracleDb_new()
//   Invoked when a new instance of the OracleDb class is created.
//-----------------------------------------------------------------------------
bool njsOracleDb_new(napi_env env, napi_value instanceObj,
        njsOracleDb **instance)
{
    njsOracleDb *oracleDb;
    napi_value temp;

    // allocate memory for structure and populate it with default values
    // memory is zero-ed so only non-zero values need to be set
    oracleDb = calloc(1, sizeof(njsOracleDb));
    if (!oracleDb) {
        njsUtils_throwError(env, errInsufficientMemory);
        return false;
    }
    oracleDb->outFormat = NJS_ROWS_ARRAY;
    oracleDb->maxRows = NJS_MAX_ROWS;
    oracleDb->stmtCacheSize = NJS_STMT_CACHE_SIZE;
    oracleDb->poolMax = NJS_POOL_MAX;
    oracleDb->poolMin = NJS_POOL_MIN;
    oracleDb->poolIncrement = NJS_POOL_INCR;
    oracleDb->poolTimeout = NJS_POOL_TIMEOUT;
    oracleDb->fetchArraySize = 100;
    oracleDb->lobPrefetchSize = NJS_LOB_PREFETCH_SIZE;
    oracleDb->poolPingInterval = NJS_POOL_DEFAULT_PING_INTERVAL;
    oracleDb->events = true;

    // wrap the structure for use by JavaScript
    if (napi_wrap(env, instanceObj, oracleDb, njsOracleDb_finalize, NULL,
            NULL) != napi_ok) {
        free(oracleDb);
        return njsUtils_genericThrowError(env);
    }

    // keep a reference to the base database object class
    NJS_CHECK_NAPI(env, napi_get_named_property(env, instanceObj,
            "BaseDbObject", &temp))
    NJS_CHECK_NAPI(env, napi_create_reference(env, temp, 1,
            &oracleDb->jsBaseDbObjectConstructor))

    // create object for storing subscriptions
    NJS_CHECK_NAPI(env, napi_create_object(env, &temp))
    NJS_CHECK_NAPI(env, napi_create_reference(env, temp, 1,
            &oracleDb->jsSubscriptions))

    *instance = oracleDb;
    return true;
}


//-----------------------------------------------------------------------------
// njsOracleDb_prepareClass()
//   Prepares the class for use by the module. This extends the prototype of
// the named class with the specified properties.
//-----------------------------------------------------------------------------
bool njsOracleDb_prepareClass(njsOracleDb *oracleDb, napi_env env,
        napi_value instance, const njsClassDef *classDef, napi_ref *clsRef)
{
    napi_value cls, prototype, tempInstance, extendFn, tempResult;
    napi_property_descriptor *allProperties, *tempProperty;
    size_t numProperties, numBaseProperties, i;

    // get the class from the instance
    NJS_CHECK_NAPI(env, napi_get_named_property(env, instance, classDef->name,
            &cls))

    // create a new instance of the class (temporarily) and get its prototype
    NJS_CHECK_NAPI(env, napi_new_instance(env, cls, 0, NULL, &tempInstance))
    NJS_CHECK_NAPI(env, napi_get_prototype(env, tempInstance, &prototype))

    // scan each of the class properties and constants to get the total number
    // of properties to define
    numProperties = 0;
    if (!classDef->propertiesOnInstance) {
        for (i = 0; classDef->properties[i].utf8name; i++, numProperties++);
    }
    numBaseProperties = numProperties;
    if (classDef->constants)
        for (i = 0; njsClassConstants[i].name; i++, numProperties++);

    // perform define if any properties are present
    if (numProperties > 0) {

        // allocate memory for all of the properties
        allProperties = calloc(numProperties,
                sizeof(napi_property_descriptor));
        if (!allProperties) {
            njsUtils_throwError(env, errInsufficientMemory);
            return false;
        }

        // populate the properties
        memcpy(allProperties, classDef->properties,
                sizeof(napi_property_descriptor) * numBaseProperties);
        if (classDef->constants) {
            for (i = 0; classDef->constants[i].name; i++) {
                tempProperty = &allProperties[numBaseProperties + i];
                tempProperty->utf8name = classDef->constants[i].name;
                if (napi_create_uint32(env, classDef->constants[i].value,
                        &tempProperty->value) != napi_ok) {
                    free(allProperties);
                    return njsUtils_genericThrowError(env);
                }
                tempProperty->attributes = napi_enumerable;
            }
        }

        // define the properties on the prototype
        if (napi_define_properties(env, prototype, numProperties,
                allProperties) != napi_ok) {
            free(allProperties);
            return njsUtils_genericThrowError(env);
        }
        free(allProperties);

    }

    // and call the _extend function defined in JavaScript
    NJS_CHECK_NAPI(env, napi_get_named_property(env, prototype, "_extend",
            &extendFn))
    NJS_CHECK_NAPI(env, napi_call_function(env, prototype, extendFn, 1,
            &instance, &tempResult))

    // keep a reference to it, if requested
    if (clsRef) {
        NJS_CHECK_NAPI(env, napi_create_reference(env, cls, 1, clsRef))
    }

    return true;
}
